/* Program to read a matrix in MM (Matrix Market) format and 
 * solve the corresponding linear system with the CGMN algorithm 
 * [Bjorck & Elfving, BIT 1979], [Gordon & Gordon, ACM TOMS 2008]. 
 * 
 * The MM format is described in 
 * https://math.nist.gov/MatrixMarket/formats.html#MMformat
 * Data types: real, complex, integer and binary ("pattern" format).
 * Matrix formats: general, symmetric, skew-symmetric and Hermitian  
 * (for complex elements). The array format is not supported. 
 *
 * For the RHS, the program also supports the array format. If the 
 * RHS is in coordinate format, the number of columns should be 1. 
 * If the matrix and/or the input RHS are complex, the system is 
 * treated as complex. 
 *
 * If an RHS vector is not provided, the program can create one by 
 * multiplying the matrix by a column of ones. In this case, the 
 * relative L_2 error (w.r.t. the vector of ones) will also be given. 
 * Alternately, one can specify the RHS to be a column of ones. 
 * 
 * To compile: gcc -O3 -o cgmn cgmn.c -lm
 * Better results may be obtained by adding the flag -ffast-math.
 * On some systems, you might also want to use the flag 
 * -Wno-unused-result to eliminate useless warnings. 
 * Usage: just run ./cgmn <enter> for detailed instructions, 
 * or see msg1[] to msg4[] below. 
 * 
 * Written by Dan Gordon, Dec. 2012.  Email for bugs, 
 * questions, experience, etc.:  gordon@cs.haifa.ac.il 
 * Minor changes and corrections: Apr. 25, 2013. 
 * Additional changes: Nov. 2015. 
 * Minor cosmetic changes: Dec. 2018. 
*/ 

#include <stdio.h>
#include <stdlib.h>
#include <string.h> 
#include <math.h>
#include <time.h>
#ifndef HUGE
  #define HUGE 3.40282347e+38F
#endif
#define MIN(a,b) (((a)<=(b)) ? (a) : (b)) 
#define MAX(a,b) (((a)>=(b)) ? (a) : (b)) 
#define SMALL 1.e-200
// SMALL is used in cgmn(...) to determine if one of the variables dot 
// or rsq is in danger of producing NAN results. SMALL can be modified 
#define LINE printf("\n") 

// IFITER defines when intermediate CGMN results are checked and printed. 
// Uncomment your preferred mode, or make up your own. 
// #define IFITER 1  
// #define IFITER (it%10==0) 
#define IFITER (it%100==0) 
// #define IFITER (it%1000==0) 
// #define IFITER ((it<=10) || (it<100 && it%10==0) ||\
//		  (it<1000 && it%100==0) || (it%1000==0)) 

// function prototypes  

void parameters(int, char**) ;
void getmat(char**) ;
void setuprhs(char**) ;
void readsoln(char**) ;
void norm_cols(char**) ;
void sparsify(char**) ; 
void norm_eqns(double) ;
long double dotc(double*, double*); 
long double dotr(double*, double*); 
void matvec(double*, double*) ;
void dkswp(double*, double*) ;
void dkswp0(double*, double*) ;
void dkswpc(double*, double*) ;
void dkswp0c(double*, double*) ;
void cgmn(double*) ;

// global variables  

char msg1[] =\
"USAGE:\n\
./cgmn matrix.mtx {rhs.mtx|RHS|RHS1} omega maxiter eps\n\
   [col|spar|col_spar|eqn_col|eqn_spar|eqn_col_spar]\n\
   [solution.mtx]\n\
matrix.mtx is an input matrix in Matrix Market format.\n\
rhs.mtx is an input RHS vector, or, if RHS is specified,\n\
   then an artificial RHS is created by multiplying\n\
   matrix.mtx by a vector of ones. In this case, the\n\
   relative L_2 error will also be given. If RHS1 is\n\
   is specified, the RHS is taken as a vector of ones.\n" ; 

char msg2[] =\
"omega is the relaxation parameter, 0 < omega < 2.\n\
maxiter is the maximum no. of iterations.\n\
eps > 0 is the convergence goal for the relative residual.\n" ; 

char msg3[] =\
"Other parameters are optional and take\n\
effect only after the RHS has been set up:\n\
col: normalizes the columns (divides each\n\
     column by the L_2-norm of its coefficients.\n\
spar: divides each column by the number of its nonzeros.\n\
col_spar: same as col followed by spar.\n" ; 

char msg4[] =\
"eqn_col, eqn_spar, eqn_col_spar: same as above,\n\
but preceded by normalizing the equations.\n\
Note that cgmn inherently normalizes the equations.\n\
solution.mtx is an optional known solution which\n\
can be given only if rhs.mtx is given.\n\
To modify the runtime printout, change the #define IFITER\n\
statement at the beginning of the program and recompile.\n" ;

int rhs=0 ;
// rhs=1 means create artificial RHS vector by 
// multiplying the matrix by a column of ones. 
int rhs1=0 ;
// rhs1=1 means create artificial RHS vector of ones. 
double omega, eps ;  // for command line parameters 
int maxiter, col=0 ; // for command line parameters 

// flags for type of matrix elements and format  
char* real=0 ; // real will become nonzero for real matrix 
char* integ=0 ;// integ will become nonzero for integer matrix  
char* gen=0 ;  // gen will become nonzero for general format  
char* sym=0 ;  // sym will become nonzero for symmetric matrix 
char* skew=0 ; // skew will become nonzero for skew-symmetric matrix 
char* comp=0 ; // comp will become nonzero for complex matrix  
char* comprhs=0 ; // comprhs will become nonzero for complex RHS  
char* compsol=0 ; // compsol will become nonzero for complex solution  
char* herm=0 ; // herm will become nonzero for Hermitian matrices  
char* pat=0 ;  // pat will become nonzero for binary pattern matrix 

int nr, nc, nnz, nnz1 ; // nr, nc are no. of rows, columns 
/* nnz is no. of nonzero elements in input matrix, nnz1 is no. of 
nonzero elements in working matrix. nnz1 differs from nnz when 
matrix is in (skew-)symmetric or Hermitian format. 
*/
int nr2, nc2 ; // twice nr and nc for complex case 
int nnr ; // nnr = comp ? nr2 : nr  
int nnc ; // nnc = comp ? nc2 : nc  
int option ; // used for determining eqn/column scaling options  
int *inrow, *incol ;      // data from input matrix 
double *inreal, *inimag ; // data from input matrix 
int *clmn ;
double *valr=0, *valc=0 ;
        // column index and values of stored matrix  
double *colnorm = 0 ; // vector for column norms 
int *nnzr, *indx ;
/* nnzr[i] will be no. of nonzeros in row i of output matrix. 
 * indx[i] will be index of first element of row i in the 
 * arrays clmn, valr, and valc. indx is also used for filling 
 * these arrays in function getmat. 
 * For convenience, all indices start at 1, as in MM format.
*/
double *bee = 0 ; // for RHS vector  
double *soln = 0 ; // for a known solution, if provided.  
int sol = 0 ;  // if not 0, solution is provided in argv[sol]  
char c, s[100] ;
FILE *fmat ; 

clock_t start;  // variable for measuring time 
double cps = (double)CLOCKS_PER_SEC;
// CLOCKS_PER_SEC is defined in time.h 

/******************************************************************/

int main(int argc, char** argv)
{ double *x ;
start = clock() ; 
parameters(argc, argv) ;
getmat(argv) ;
setuprhs(argv) ; 
if (sol) readsoln(argv) ; 

if (option) switch(option) 
   { case 1: norm_cols(argv) ; break ; 
     case 2: sparsify(argv) ; break ; 
     case 3: norm_cols(argv) ; sparsify(argv) ; break ; 
     case 4: norm_eqns(1.0) ; norm_cols(argv) ; break ; 
     case 5: norm_eqns(1.0) ; sparsify(argv) ; break ; 
     case 6: norm_eqns(1.0) ; norm_cols(argv) ; sparsify(argv) ; 
   }
     
if (comp) x = (double*)calloc(nc2+1, sizeof(double)) ;
else x = (double*)calloc(nc+1, sizeof(double)) ;
if (!x) { printf("\nCould not allocate memory for vector x in main().\n") ; 
	  exit(1) ; 
	} 

cgmn(x) ;

free(indx) ;
free(clmn) ;
free(bee) ;
free(valr) ;
if (comp) free(valc) ;
if (col)  free(colnorm) ;

/* Computed solution is in x[1],...,x[nnc]. If the system is 
 * complex, the solution is x[1]+i*x[2],..., x[nnc-1]+i*x[nnc]. 
 * If a solution vector was provided, the values are in array 
 * soln, ordered similarly to x. 
 * Insert code here for additional processing of x or soln. */

free(x) ; 
if (sol) free(soln) ; 

} // end of main  

/*******************************************************************/

void parameters(int argc, char **argv)
{ int k ;
struct tm *local;
time_t t;

if (argc<6)
   { printf("\n%s%s%s%s\n\n", msg1, msg2, msg3, msg4); exit(1); }

t = time(NULL);
local = localtime(&t);
printf("\n============================================================") ;
printf("\nDate & time: %s", asctime(local));

printf("\nExecuting: ") ; 
for(k=0; k<argc; k++) printf("%s ", argv[k]) ; 

k = argc ;
if (k > 8) 
   { printf("\nExtra parameters ignored.") ; 
     k = 8 ; 
   } 

if (!strcmp(argv[2], "RHS1")) rhs1 = 1 ;
else if (!strcmp(argv[2], "RHS")) rhs = 1 ;

omega = atof(argv[3]) ;
if (omega<=0.||omega>=2.) 
   { printf("\nomega should be  0 < omega < 2.\n\n") ; 
     exit(0) ;
   } 
maxiter = atoi(argv[4]) ;
if (maxiter <= 0) 
   { printf("\nNo. of iterations should be > 0.\n\n") ; 
     exit(0) ;
   }
eps = atof(argv[5]) ;
if (eps<=0.)
   { printf("\neps should be > 0.\n\n") ;
     exit(0) ;
   }
if (rhs)
   printf("\n\nRunning CGMN on matrix %s. RHS created\n\
by multiplying matrix by column of ones.\n", argv[1]) ;
else if (rhs1)
   printf("\n\nRunning CGMN on matrix %s. \n\
RHS created as a column of ones.\n", argv[1]) ;
else
   printf("\n\nRunning CGMN on matrix %s. RHS is %s\n", argv[1], argv[2]) ;

printf("Relax. param.: %g  max-iter.: %d  rel-res goal: %g\n",\
omega, maxiter, eps) ;

/* Set options for various scaling operations 
 * and optional solution vector. */ 
option = 0 ; 
sol = 0 ; 
if (k>=7) 
   { if (!strcmp(argv[6],"col")) option = 1 ; 
     else if (!strcmp(argv[6],"spar")) option = 2 ; 
     else if (!strcmp(argv[6],"col_spar")) option = 3 ; 
     else if (!strcmp(argv[6],"eqn_col")) option = 4 ; 
     else if (!strcmp(argv[6],"eqn_spar")) option = 5 ; 
     else if (!strcmp(argv[6],"eqn_col_spar")) option = 6 ; 
   }

if (k>=7 && !option) sol = 6 ;
if (k==8 &&  option) sol = 7 ;
if (sol && (rhs||rhs1)) // consistency check 
   { printf("\nSolution vector cannot be used if RHS or RHS1\n\
are specified. Ignoring parameter %s\n", argv[sol]) ;
     sol = 0 ;
   }
if (sol) printf("\n%s is used as the solution vector\n", argv[sol]);
if (k==8 && !option) printf("Extra parameter %s is ignored.\n",argv[7]) ;
} // end of parameters  

/**********************************************************************/

void getmat(char **argv)
{ int i, k, nd=0 ;
// nd is counter of nonzero diagonal elements 
double sumdiag = 0., sumoff = 0. ;
// sums of diag. and off-diag. elements 
char *mm=0, *coord=0 ;

int minnz, maxnz ; // min of no. of nonzeros in a row 
double avgnz, varnz, stdnz ; // avg, var, std of nonzeros in a row 

printf("\nReading and setting up matrix...\n") ;
fmat  = fopen(argv[1], "r") ;
if (!fmat)
   { printf("\nERROR: could not open input file %s\n\n", argv[1]) ;
     exit(1) ;
   }

k = 1 ; // flag for first line 
c = fgetc(fmat) ;
while (c=='%')
   { ungetc(c, fmat) ; // restore first character 
     fgets(s, 90, fmat) ;
     if (k) // first line 
        { mm = strstr(s, "MatrixMarket") ;
	  coord = strstr(s, "coordinate") ;
	  if (!mm||!coord) 
	     { printf("\nInput matrix is not in MatrixMarket format\n\
or it is not in coordinate format.\n\n") ;
	       exit(1) ;
	     }
	  real = strstr(s, "real") ; 
	  integ = strstr(s, "integer") ; 
          comp = strstr(s, "complex") ;
	  gen = strstr(s, "general") ; 
	  sym  = strstr(s, "symmetric") ;
	  herm = strstr(s, "ermitian") ; // could be h or H...
	  if (sym) skew = strstr(s, "skew") ; 
	  pat = strstr(s, "pattern") ; 
          k = 0 ;
        }
     c = fgetc(fmat) ;
   }
ungetc(c, fmat) ;

fscanf(fmat, "%d %d %d", &nr, &nc, &nnz) ;
nr2=nr+nr; 
nc2=nc+nc; 
nnr = comp ? nr2 : nr ; 
nnc = comp ? nc2 : nc ; 
inrow = (int*)calloc((nnz+1),sizeof(int)) ;
incol = (int*)calloc((nnz+1),sizeof(int)) ;
inreal = (double*)calloc((nnz+1),sizeof(double)) ;
if (!inrow||!incol||!inreal)
   { printf("\nCould not allocate memory for input matrix\n") ;
     exit(2) ;
   }
if (comp) 
   { inimag = (double*)calloc((nnz+1),sizeof(double)) ;
     if (!inimag)
        { printf("\nCould not allocate memory for input matrix\n") ;
          exit(2) ;
        }
   }

// nnzr is array for no. of nonzeros in each row 
// A complex number is counted as one nonzero  

nnzr = (int*)calloc(nr+1, sizeof(int)) ;
if (!nnzr) 
   { printf("\nCould not allocate memory for array nnzr\n") ;
     exit(2) ;
   }

// read input elements: 
for (k=1; k<=nnz; k++)
   if (pat)  // just 0-1 matrix 
      { fscanf(fmat, "%d %d", &inrow[k], &incol[k]) ;
	inreal[k] = 1.0 ;
      } 
   else 
      if (comp)  // complex matrix  
         fscanf(fmat,"%d %d %lg %lg", &inrow[k], &incol[k],\
	 &inreal[k], &inimag[k]) ; 
      else  // real matrix  
         fscanf(fmat,"%d %d %lg",&inrow[k], &incol[k], &inreal[k]) ;

fclose(fmat) ;

nnz1 = nnz ; // # nonzeros - will be modified if sym or herm 
for (k=1; k<=nnz; k++)
   { i = inrow[k] ;
     nnzr[i]++ ; // increment no. of nonzeroes of row i 
     if (i==incol[k]) // diag. element  
        { sumdiag += fabs(inreal[k]) ;
          nd++ ; 
        } 
     else // off-diag. element 
	{ sumoff += fabs(inreal[k]) ;
	  if (sym||herm) // make room for "symmetric" element 
             nnzr[incol[k]]++ ; 
	}
   }
if (skew && !nd) // consistency check  
   { printf("\nERROR: matrix is defined as skew-symmetric, but there\n\
are elements on the diagonal.  Terminating program.\n\n"); 
     exit(1) ; 
   } 

// modify # nonzeros for sym or herm matrices  
if (sym||herm) nnz1 += nnz - nd ; // for skew, nd is 0  

// check that every row has at least one element in input matrix: 
for (i=1; i<=nr; i++)
if (nnzr[i]==0) 
   { printf("\nERROR: row %d has no elements in input matrix.\n\
Aborting program.\n", i) ; 
     exit(9) ;
   }

//statistics for number of nonzeros in each row: 
avgnz = varnz = 0 ; 
minnz = maxnz = nnzr[1] ;
for (i=1; i<=nr; i++)
   { avgnz += nnzr[i] ;
     varnz += nnzr[i]*nnzr[i] ;
     if (nnzr[i] < minnz) minnz = nnzr[i] ;
     if (nnzr[i] > maxnz) maxnz = nnzr[i] ; 
   }
avgnz /= nr ;
varnz = varnz/nr - avgnz*avgnz ;
stdnz = sqrt(varnz) ;
printf("\nStatistics for number of nonzeros in each row:\n\
min = %d  max = %d  avg = %g  var = %g  std = %g\n",\
minnz, maxnz, avgnz, varnz, stdnz) ; 

clmn = (int*)calloc((nnz1+1),sizeof(int)) ;
valr = (double*)calloc((nnz1+1),sizeof(double)) ;
if (!clmn||!valr)
   { printf("\nCould not allocate memory for matrix data\n") ;
     exit(2) ;
   }
if (comp) 
   { valc = (double*)calloc((nnz1+1),sizeof(double)) ; 
     if (!valc) 
        { printf("\nCould not allocate memory for matrix data\n") ;
          exit(2) ;
        } 
   }

indx = (int*)malloc((nr+2)*sizeof(int)) ;
// extra element in indx to simplify coding 
if (!indx) // indx[i]=position of start of row i in clmn and valr 
   { printf("\nCould not allocate memory for array indx\n") ;
     exit(2) ;
   }

/* While filling valr, valc and clmn, indx[i] is used to point 
 * to the next available position of a nonzero element of row i. 
 * After this fill, indx[i] is set to point to the first element 
 * of row i in valr, valc and clmn. 
 */ 

indx[1] = 1 ;
for (i=2; i<=nr; i++)
   indx[i] = indx[i-1] + nnzr[i-1] ;

// copy input elements to internal matrix representation: 
for (k=1; k<=nnz; k++)
   { i = indx[inrow[k]] ;
     clmn[i] = incol[k] ;
     valr[i] = inreal[k] ;
     if (comp) valc[i] = inimag[k] ; 
     indx[inrow[k]]++ ;
// if symmetric or Hermitian format, add required element 
     if ((sym||herm) && (inrow[k] != incol[k]))
        { i = indx[incol[k]] ;
          clmn[i] = inrow[k] ;
	  if (skew) valr[i] = -inreal[k] ; 
	  else valr[i] = inreal[k] ;
          if (comp)  // complex elements  
	     if (skew||herm) valc[i] = -inimag[k] ; 
	     else valc[i] = inimag[k] ; 
          indx[incol[k]]++ ;
        }
   }

/* set indx[i] to point to position of first element 
of row i in arrays valr, valc and clmn. Current values 
of indx are modified in reverse order. Loop starts with 
k=nr+1 to simplify coding when scanning last row. */ 

for (k=nr+1; k>1; k--) 
   indx[k] = indx[k-1] ; 
// indx[k-1] has already advanced to where indx[k] should point 
indx[1] = 1 ; 

free(inrow) ;
free(incol) ;
free(inreal) ;
if (comp) free (inimag) ; 
free(nnzr) ;

printf("Matrix contains ") ;

if (real) printf("real elements in ") ;
if (comp) printf("complex elements in ") ;
if (integ) printf("integer elements in ") ;
if (pat) printf("binary elements in ") ;

if (gen) printf("general format.") ;
if (sym && !skew) printf("symmetric format.") ;
if (skew) printf("skew-symmetric format.") ;
if (herm) printf("Hermitian format.") ;

printf("\n#rows: %d  #cols: %d  #nonzeros: %d\n", nr, nc, nnz1) ;

if (!comp) 
{ sumdiag = (nd) ? sumdiag/MIN(nc,nr) : 0. ;
  printf("Avg diag: %g  avg off-diag: %g\n", sumdiag, sumoff/(nnz-nd));
}

} // end of getmat  

/***************************************************************/

void setuprhs(char **argv) 
// function to set up the RHS  
{ 
int i, ii, j, k, n, nz ;
double valreal, valcomp ; 
char *mm=0, *coord=0 ; 

printf("\nSetting up RHS...\n") ;
bee = (double*)calloc((nnr+1), sizeof(double)) ; 
if (!bee)
   { printf("\nERROR: could not allocate memory for bee\n") ;
     exit(1) ;
   }

if (rhs)  // create RHS by multiplying matrix by vector of ones 
   { soln = (double*)calloc((nnc+1), sizeof(double)) ;
     if (!soln)
        { printf("\nERROR: could not allocate memory for soln\n") ;
          exit(1) ;
        }
     for (k=1; k<=nnc; k++) soln[k] = 1.0 ; 
     if (comp)  // complex matrix 
        for (i=1; i<=nr; i++)
           for (k=indx[i]; k<indx[i+1]; k++)
              { bee[2*i-1] += valr[k] - valc[k] ; // real part of bee  
	        bee[2*i]   += valr[k] + valc[k] ; // imag part of bee  
	      } 
     else  // real matrix  
        for (i=1; i<=nr; i++) 
           for (k=indx[i]; k<indx[i+1]; k++)
              bee[i] += valr[k] ;
   }
else if (rhs1)  // RHS = vector of ones 
        for (i=1; i<=nnr; i++) 
	   bee[i] = 1.0 ;

else // read RHS from input file 
   { fmat  = fopen(argv[2], "r") ;
     if (!fmat)
        { printf("\nERROR: could not open input file %s\n\n", argv[2]) ;
          exit(1) ;
        }
     c = fgetc(fmat) ;
     k = 1 ; // flag for first line  
     while (c=='%')
        { ungetc(c, fmat) ; // restore first character 
          fgets(s, 90, fmat) ;
	  if (k) // first line  
	     { mm = strstr(s, "MatrixMarket") ;
	       if (!mm) 
		  { printf("\nRHS vector is not in MatrixMarket format\n");
		    exit(1) ;
		  }
	       coord = strstr(s, "coordinate") ;
	       comprhs = strstr(s, "complex") ;
	       k = 0 ;
	     }
	  c = fgetc(fmat) ;
        }
     ungetc(c, fmat) ;

     // read size of RHS: 
     if (coord)  // coordinate format  
        fscanf(fmat, "%d %d %d", &n, &k, &nz) ; 
     else  // array format  
        fscanf(fmat, "%d %d", &n, &k) ;

     // consistency check: 
     if (n!=nr) 
        { printf("\nERROR: # matrix rows != size of RHS vector\n") ;
          exit(1) ;
        }

     if (coord && k>1)  // RHS matrix has k>1 columns  
	{ j = k + 1 ; 
	  while (j>k)
	     { printf("\nRHS matrix %s has %d columns.\n\
Enter col. number (between 1 and %d)\n\
to continue (0 to exit): ", argv[2], k, k) ; 
	       scanf("%d", &j) ;
               if (j<=0) exit(0) ;
	       if (j>k) 
	          printf("%d is too large. Try again...", j) ;
	     }
	}
     else j = 1 ;  // for matrices with one column  
     if (comp && !comprhs) 
	printf("\nNote: matrix is complex and RHS is real. System\n\
is treated as complex by adding imaginary zeros to RHS.\n") ;

     if (!comp && comprhs)  // not very likely, but still...  
	{ printf("\nNote: matrix is real and RHS is complex. System\n\
is treated as complex by adding imaginary zeros to matrix.\n") ; 
	  comp = "complex" ;  // system is treated as complex  
	  nnr = nr2 ; 
	  nnc = nc2 ; 
	  valc = (double*)calloc((nnz1+1), sizeof(double)) ;
          if (!valc)
             { printf("\nCould not allocate memory for matrix data\n") ;
               exit(2) ;
             }
	  free(bee) ; // this is preferable to realloc  
	  bee = (double*)calloc(nnr+1, sizeof(double)) ; 
	  if (!bee)
	     { printf("\nERROR: could not allocate memory for bee\n") ;
	       exit(1) ;
	     }	
	 }

     // read RHS into bee:  

     if (coord)  // coordinate format 
        for (i=1; i<=nz; i++)
	   if (comprhs)  // complex RHS  
	      { fscanf(fmat, "%d %d %lg %lg", &n, &k, &valreal, &valcomp) ;
		if (k==j) 
                   { bee[2*n-1] = valreal ;
                     bee[2*n]   = valcomp ;
		   }
              }
           else   // real RHS 
              { fscanf(fmat, "%d %d %lg", &n, &k, &valreal) ;
	        if (k==j) 
                   if (comp)  // matrix is complex, RHS is real 
                      bee[2*n-1] = valreal ; // bee[2n] is already 0 
                   else  // matrix and RHS are real  
		      bee[n] = valreal ;
              }

     else  // array format  
        for (i=1; i<=nnr; i++)
	   if (comprhs)  // complex RHS 
              fscanf(fmat, "%lg %lg", &bee[2*i-1], &bee[2*i]) ;
           else  // real RHS 
              if (comp)  // matrix is complex, RHS is real 
                 fscanf(fmat, "%lg", &bee[2*i-1]) ; // bee[2i] is already 0 
              else  // matrix and RHS are real  
		 fscanf(fmat, "%lg", &bee[i]) ;

     fclose(fmat) ;
   }

// check that bee is not identically 0: 

i = 1 ; 
while (i<=nnr && bee[i]==0) i++ ; 
if ( i > nnr ) 
   { printf("\nRHS consists only of zeros, so solution\n\
is zero. Exiting program.\n\n") ;
     exit(4) ;
   }

} // end of setuprhs 

/***************************************************************/

void readsoln(char **argv) 
// function to read a solution vector  
{ 
int i, ii, j, k, n, nz ;
double valreal, valcomp, sum1, sum2 ; 
double* x ; 
char *mm=0, *coord=0 ; 

printf("\nReading the solution vector...\n") ;
soln = (double*)calloc((nnc+1), sizeof(double)) ; 
if (!soln)
   { printf("\nERROR: could not allocate memory for soln\n") ;
     exit(1) ;
   }

fmat  = fopen(argv[sol], "r") ; // sol is 7 or 8  
if (!fmat)
   { printf("\nERROR: could not open input file %s\n\n", argv[sol]) ;
     exit(1) ;
   }
c = fgetc(fmat) ;
k = 1 ; // flag for first line  
while (c=='%')
   { ungetc(c, fmat) ; // restore first character 
     fgets(s, 90, fmat) ;
     if (k) // first line  
	{ mm = strstr(s, "MatrixMarket") ;
	  if (!mm) 
	     { printf("\nERROR: solution vector is not in\n\
MatrixMarket format. Terminating program.\n");
	       exit(1) ;
	     }
	  coord = strstr(s, "coordinate") ;
	  compsol = strstr(s, "complex") ;
	  k = 0 ;
	}
     c = fgetc(fmat) ;
   }
ungetc(c, fmat) ;

if (coord) // only the array format is supported for the solution vector  
   { printf("\nERROR: the solution vector is in coordinate\n\
format, but only the array format is supported\n\
for the solution vector. Terminating program.\n") ; 
     exit(3) ; 
   } 

fscanf(fmat, "%d %d", &n, &k) ;

     // consistency checks: 
if (n!=nc) 
   { printf("\nERROR: size of solution vector != # of columns in matrix.\n") ;
     exit(1) ;
   }
if (k > 1) 
   { printf("\nERROR: solution vector has %d columns, but only one\n\
column is allowed for such vectors. Terminating program.\n", k) ;
     exit(3) ; 
   }
if (!comp && !comprhs && compsol)  // impossible!  
   { printf("\nERROR: matrix and RHS are real, but the input\n\
solution is complex; this is impossible.\n") ; 
     exit(1) ; 
   } 

// read solution vector into soln:  

for (i=1; i<=nc; i++)
   if (compsol)  // complex solution 
      fscanf(fmat, "%lg %lg", &soln[2*i-1], &soln[2*i]) ;
   else  // real solution 
      if (comp || comprhs)  // matrix or rhs complex, solution is real 
         { // the solution vector must be treated as complex  
           fscanf(fmat, "%lg", &soln[2*i-1]) ; // soln[2i] is already 0 
	   compsol = "complex" ; 
         }
      else  // matrix and solution are real  
         fscanf(fmat, "%lg", &soln[i]) ;
fclose(fmat) ;

// verify that soln is actually the solution  

x = (double*)calloc((nnr+1), sizeof(double)) ; 
if (!x) 
   { printf("\nCould not allocate space for vector x in readsoln.\n") ; 
     exit(1) ; 
   } 
matvec(soln, x) ;  // x = mat*soln  
sum1 = sum2 = 0. ; 
for (k=1; k<=nnr; k++) 
   { sum1 += (x[k] - bee[k])*(x[k] - bee[k]) ; 
     sum2 += bee[k]*bee[k] ; 
   } 
printf("Verification of the input solution vector:\n\
rel-err = ||mat*soln - b||/||b|| = %lg\n", sqrt(sum1/sum2)) ; 
free(x) ; 

} // end of readsoln 

/***************************************************************/

void norm_cols(char **argv) 
{ int j, k ; 
double min, max, avg, var ; 
double std, pcstd[10], length ; 
char name[80] ; 
int hist[100] ; 

col = 1 ; 
colnorm = (double*)calloc(nnc+1, sizeof(double)) ;
if (!colnorm)
   { printf("\nERROR: could not allocate memory for colnorm\n\n");
     exit(1) ;
   }

printf("\nNormalizing columns before CGMN by dividing each\n\
column by the L_2 norm of its coefficients...\n") ;
if (comp) // complex case  
   for (k=1; k<=nnz1; k++)
      { j = clmn[k] ; 
        colnorm[j] += valr[k]*valr[k] + valc[k]*valc[k] ;
      }
else  // real case  
   for (k=1; k<=nnz1; k++)
      colnorm[clmn[k]] += valr[k]*valr[k] ;

for (k=1; k<=nnc; k++)
   { if (colnorm[k]==0.0)
        { printf("\n***** WARNING! ***** \n\n\
All elements in column %d of input matrix are zero.\n\
Program continues without normalizing the columns.\n", k) ;
          free(colnorm) ; 
          col = 0 ; 
          return ;
        }
     colnorm[k] = sqrt(colnorm[k]) ; 
   }

for (k=1; k<=nnz1; k++)
   { valr[k] /= colnorm[clmn[k]] ;
     if (comp) valc[k] /= colnorm[clmn[k]] ; 
   }

// double colnorm for complex case: 
if (comp) for(j=nnc; j>=2; j-=2)
   colnorm[j] = colnorm[j-1] = colnorm[j/2] ;

// calculate statistics for colnorm:  
avg = var = 0. ;
min = max = colnorm[1] ;
for (j=1; j<=nnc; j++) 
   { avg += colnorm[j] ;
     var += colnorm[j]*colnorm[j] ;
     if (colnorm[j] < min) min = colnorm[j] ;
     if (colnorm[j] > max) max = colnorm[j] ;
   } 
avg /= nnc ;
var = var/nnc - avg*avg ;
std = sqrt(var) ; 
printf("Statistics for column norms:\n\
min = %g  max = %g  avg = %g  var = %g  std = %g\n",\
min, max, avg, var, std) ;

} // end of norm_cols 

/***************************************************************/

void sparsify(char **argv) 
// divide each column by the number of its nonzeros  
{ int j, k ; 
double min, max, avg, var ; 
double std, pcstd[10], length ; 
double * spvec ; 
// spvec[j] will hold the number of nonzeros in column j  
char name[80] ; 
int hist[100] ; 

col = 1 ; 
if (comp) spvec = (double*)calloc(nnc+1, sizeof(double)) ;
else spvec = (double*)calloc(nc+1, sizeof(double)) ;
if (!spvec)
   { printf("\nERROR: could not allocate memory for spvec\n\n");
     exit(1) ;
   }

printf("\nDividing each column by the number of its nonzeros ...\n") ;
for (k=1; k<=nnz1; k++)
   if (valr[k]!=0||(comp&&(valc[k]!=0))) spvec[clmn[k]] += 1.0 ; 

// EXPERIMENT WITH sqrt(spvec)  
// for (k=1; k<=nc; k++) spvec[k] = sqrt(spvec[k]) ;  

for (k=1; k<=nc; k++)
   { if (spvec[k]==0.0)
        { printf("\n***** WARNING! ***** \n\n\
All elements in column k=%d, clmn[k]=%d of matrix are zero.\n\
spvec[k] = %g.\n\
Program continues without sparsifing the columns.\n", k,clmn[k],spvec[k]) ;
          free(spvec) ; 
          col = 0 ; 
          return ;
        }
   }

for (k=1; k<=nnz1; k++)
   { valr[k] /= spvec[clmn[k]] ;
     if (comp) valc[k] /= spvec[clmn[k]] ; 
   }

// double spvec for complex case:  
   if (comp) for (j=nnc; j>=2; j-=2)
      spvec[j] = spvec[j-1] = spvec[j/2] ;

// calculate statistics for spvec:  
avg = var = 0. ;
min = max = spvec[1] ;
for (j=1; j<=nnc; j++) 
   { avg += spvec[j] ;
     var += spvec[j]*spvec[j] ;
     if (spvec[j] < min) min = spvec[j] ;
     if (spvec[j] > max) max = spvec[j] ;
   } 
avg /= nnc ;
var = var/nnc - avg*avg ;
std = sqrt(var) ; 
printf("Statistics for sparsity of columnns:\n\
min = %g  max = %g  avg = %g  var = %g  std = %g\n",\
min, max, avg, var, std) ;

// If columns were normalized, multiply colnorm by spvec:  
if (colnorm) 
   { for (k=1; k<=nnc; k++) colnorm[k] *= spvec[k] ; 
     free(spvec);  // don't need it any more 
   } 
else // set colnorm to be identical to spvec  
   colnorm = spvec ; 

} // end of sparsify 

/***************************************************************/

void matvec(double* x, double* y)
// Sets y = A*x (A is global).  
{ int i, j, k ; 
  for (i=1; i<=nnc; i++) y[i] = 0. ;
  if (comp)  // complex case 
     for (i=1; i<=nc; i++) 
	{ y[2*i-1] = y[2*i] = 0. ; 
	  for (k=indx[i]; k<indx[i+1]; k++) 
	     { j = clmn[k] ; 
	       y[2*i-1] += valr[k]*x[2*j-1] - valc[k]*x[2*j] ; 
	       y[2*i]   += valc[k]*x[2*j-1] + valr[k]*x[2*j] ;
	     } 
	} 
  else  // real case 
  for (i=1; i<=nnc; i++) 
     { y[i] = 0. ; 
       for (k=indx[i]; k<indx[i+1]; k++)
          y[i] += valr[k] * x[clmn[k]] ; 
     } 
} // end of matvec  

/***************************************************************/

void norm_eqns(double omeg)
/* This function normalizes the equations. 
 * Normalization is done by dividing each equation by its 
 * L_2 norm.  The equations are also multiplied by the square 
 * root of the parameter omeg in order to avoid this operation in the 
 * Kaczmarz loop.  Hence, if called from the function cgmn, omeg should 
 * the specified omega, and if called just to normalize the equations, 
 * omeg should be 1.
*/
{ int i, k ;
  double factor ;
  double min, max, avg, var, sqromeg ; 

  min=HUGE; max=-1.; avg=0.; var=0. ; 
  sqromeg = sqrt(omeg) ; 
  printf("\nNormalizing equations...\n") ;
  if (comp) // complex case 
     for (i=1; i<=nr; i++)
        { factor = 0. ;
          for (k=indx[i]; k<indx[i+1]; k++)
             factor += valr[k]*valr[k] + valc[k]*valc[k] ;

if (factor==0.0) 
   { printf("\nAll elements in row %d of input matrix are zero.\n\
Aborting program.\n", i) ;
     exit(9) ;
   }

          factor = sqrt(factor) ;
	  avg += factor ; 
	  var += factor*factor ;
	  if (factor<min) min = factor ; 
	  if (factor>max) max = factor ; 
          factor = sqromeg/factor ;
          bee[2*i-1] *= factor ;
	  bee[2*i]   *= factor ;
          for (k=indx[i]; k<indx[i+1]; k++)
             { valr[k] *= factor ; valc[k] *= factor ; }
        }

  else // real case  
     for (i=1; i<=nr; i++)
        { factor = 0. ;
          for (k=indx[i]; k<indx[i+1]; k++)
             factor += valr[k] * valr[k] ; 

if (factor==0.0)
   { printf("\nAll elements in row %d of input matrix are zero.\n\
Aborting program.\n", i) ;
     exit(9) ;
   }

	  factor = sqrt(factor) ;
          avg += factor ;
          var += factor*factor ;
          if (factor<min) min = factor ;
          if (factor>max) max = factor ;
          factor = sqromeg/factor ;
          bee[i] *= factor ;
          for (k=indx[i]; k<indx[i+1]; k++)
             valr[k] *= factor ;
        }

// calculate statistics for row norms: 
avg /= nc ;
var = var/nc - avg*avg ;
printf("Statistics for row norms:\n\
min = %g  max = %g  avg = %g  var = %g  std = %g\n",\
min, max, avg, var, sqrt(var)) ; 

} // end of norm-eqns  

/***************************************************************/

void dkswp(double* x, double* y)
// sets y = DKSWP(A, b, x, omega)  
// omega already taken care of by norm_eqns 
{ int i, k ;
  double aiy ;

  for (k=1; k<=nc; k++) y[k] = x[k] ;
// Forward sweep: 
  for (i=1; i<=nr; i++)
     { aiy = 0. ; // aiy = <a_i, y> 
       for (k=indx[i]; k<indx[i+1]; k++)
	  aiy += valr[k] * y[clmn[k]] ;
       for (k=indx[i]; k<indx[i+1]; k++)
	  y[clmn[k]] += (bee[i]-aiy) * valr[k] ;
     }
// Backward sweep: 
  for (i=nr; i>=1; i--)
     { aiy = 0. ;
       for (k=indx[i]; k<indx[i+1]; k++)
          aiy += valr[k] * y[clmn[k]] ;
       for (k=indx[i]; k<indx[i+1]; k++)
          y[clmn[k]] += (bee[i]-aiy) * valr[k] ;
     }
} // end of dkswp  

/***************************************************************/

void dkswp0(double* x, double* y)
// sets y = DKSWP(A, b=0, x, omega) 
// omega already taken care of by norm_equations 
{ int i, k ;
  double aiy ;

  for (k=1; k<=nc; k++) y[k] = x[k] ;
// Forward sweep: 
  for (i=1; i<=nr; i++)
     { aiy = 0. ;
       for (k=indx[i]; k<indx[i+1]; k++)
          aiy += valr[k] * y[clmn[k]] ;
       for (k=indx[i]; k<indx[i+1]; k++)
          y[clmn[k]] -= aiy * valr[k] ;
     }
// Backward sweep: 
  for (i=nr; i>=1; i--)
     { aiy = 0. ;
       for (k=indx[i]; k<indx[i+1]; k++)
          aiy += valr[k] * y[clmn[k]] ;
       for (k=indx[i]; k<indx[i+1]; k++)
          y[clmn[k]] -= aiy * valr[k] ;
     }
} // end of dkswp0 

/***************************************************************/

void dkswpc(double* x, double* y)
// Same as dkswp, but for complex case 
{ int i, j, k ;
  double aiy ;

  for (k=1; k<=nc2; k++) y[k] = x[k] ;
// Forward sweep: 
  for (i=1; i<=nr; i++)
     { aiy = 0. ; // for eq. 2i-1 
       for (k=indx[i]; k<indx[i+1]; k++)
	  { j = clmn[k] ;
            aiy += valr[k]*y[2*j-1] - valc[k]*y[2*j] ;
	  }
       for (k=indx[i]; k<indx[i+1]; k++)
          { j = clmn[k] ;
            y[2*j-1] += (bee[2*i-1]-aiy) * valr[k] ;
	    y[2*j]   += (bee[2*i-1]-aiy) * (-valc[k]) ;
	  }
       aiy = 0. ; // for eq. 2i 
       for (k=indx[i]; k<indx[i+1]; k++) 
	  { j = clmn[k] ; 
	    aiy += valc[k]*y[2*j-1] + valr[k]*y[2*j] ;
	  }
       for (k=indx[i]; k<indx[i+1]; k++)
	  { j = clmn[k] ;
	    y[2*j-1] += (bee[2*i]-aiy) * valc[k] ;
	    y[2*j]   += (bee[2*i]-aiy) * valr[k] ;
	  }
     }
// Backward sweep: 
  for (i=nr; i>=1; i--)
     { aiy = 0. ; // for eq. 2i 
       for (k=indx[i]; k<indx[i+1]; k++)
          { j = clmn[k] ;
            aiy += valc[k]*y[2*j-1] + valr[k] *y[2*j] ;
          }
       for (k=indx[i]; k<indx[i+1]; k++)
          { j = clmn[k] ;
            y[2*j-1] += (bee[2*i]-aiy) * valc[k] ; 
            y[2*j]   += (bee[2*i]-aiy) * valr[k]  ;
          }      
       aiy = 0. ; // for eq. 2i-1  
       for (k=indx[i]; k<indx[i+1]; k++) 
	  { j = clmn[k] ;
	    aiy += valr[k]*y[2*j-1] - valc[k]*y[2*j] ;
	  }
       for (k=indx[i]; k<indx[i+1]; k++)
	  { j = clmn[k] ; 
	    y[2*j-1] += (bee[2*i-1]-aiy) * valr[k] ;
            y[2*j]   += (bee[2*i-1]-aiy) * (-valc[k]) ; 
	  }
     }
} // end of dkswpc 

/***************************************************************/

void dkswp0c(double* x, double* y)
// Same as dkswp0, but for complex case 
{ int i, j, k ;
  double aiy ;

  for (k=1; k<=nc2; k++) y[k] = x[k] ;
// Forward sweep: 
  for (i=1; i<=nr; i++)
     { aiy = 0. ; // for eq. 2i-1 
       for (k=indx[i]; k<indx[i+1]; k++)
	  { j = clmn[k] ;
            aiy += valr[k]*y[2*j-1] - valc[k]*y[2*j] ;
	  }
       for (k=indx[i]; k<indx[i+1]; k++)
          { j = clmn[k] ;
            y[2*j-1] -= aiy * valr[k] ;
	    y[2*j]   += aiy * valc[k] ;
	  }
       aiy = 0. ; // for eq. 2i 
       for (k=indx[i]; k<indx[i+1]; k++) 
	  { j = clmn[k] ; 
	    aiy += valc[k]*y[2*j-1] + valr[k]*y[2*j] ;
	  }
       for (k=indx[i]; k<indx[i+1]; k++)
	  { j = clmn[k] ;
	    y[2*j-1] -= aiy * valc[k] ;
	    y[2*j]   -= aiy * valr[k] ;
	  }
     }
// Backward sweep: 
  for (i=nr; i>=1; i--)
     { aiy = 0. ; // for eq. 2i 
       for (k=indx[i]; k<indx[i+1]; k++)
          { j = clmn[k] ;
            aiy += valc[k]*y[2*j-1] + valr[k] *y[2*j] ;
          }
       for (k=indx[i]; k<indx[i+1]; k++)
          { j = clmn[k] ;
            y[2*j-1] -= aiy * valc[k] ; 
            y[2*j]   -= aiy * valr[k]  ;
          }      
       aiy = 0. ; // for eq. 2i-1  
       for (k=indx[i]; k<indx[i+1]; k++) 
	  { j = clmn[k] ;
	    aiy += valr[k]*y[2*j-1] - valc[k]*y[2*j] ;
	  }
       for (k=indx[i]; k<indx[i+1]; k++)
	  { j = clmn[k] ; 
	    y[2*j-1] -= aiy * valr[k] ;
            y[2*j]   += aiy * valc[k] ; 
	  }
     } 
} // end of dkswp0c 

/***************************************************************/

long double dotc(double* x, double* y)
// Returns the dot product of vectors x and y of length nc(2) 
{ int j ; 
  long double sum = 0. ;
  for (j=1; j<=nnc; j++)
     sum += x[j] * y[j] ;
  return sum ;
}

long double dotr(double* x, double* y)
// Returns the dot product of vectors x and y of length nr(2) 
{ int j ; 
  long double sum = 0. ;
  for (j=1; j<=nnr; j++)
     sum += x[j] * y[j] ;
  return sum ;
}

/***************************************************************/

void cgmn(double* x)
/* Applies CGMN to the system Ax=b, using relaxation 
 * parameter omega, for iter iterations, or until 
 * rel-res < eps. The initial estimate is x, which 
 * is set as zero in main(), and it is also the result.
*/
{ 
int j, it, indres=0, inderr=0 ;
long double alpha, beta, rsq, rsq1, dot ;
long double minb, maxb, absb ; 
double *p, *q, *r, *y ;
long double resid, relres, relerr, err, maxres, maxerr, normbee, normsol ; 
double timet ;

printf("\nStarting CGMN") ; 
p = (double*)malloc((nnr+1)*sizeof(double)) ;
q = (double*)malloc((nnr+1)*sizeof(double)) ;
r = (double*)malloc((nnr+1)*sizeof(double)) ;
y = (double*)malloc((nnc+1)*sizeof(double)) ;
if (!p||!q||!r||!y) 
   { printf("\nERROR: could not allocate space for cgmn\n") ; 
     exit(2) ;
   }

norm_eqns(omega) ; // also takes care of omega 
normbee = sqrt(dotr(bee,bee)) ;
minb = HUGE ; 
maxb = -1. ; 
for (j=1; j<=nnc; j++)
   { absb = fabs(bee[j]) ; 
     if (absb > maxb) maxb = absb ;
     if (absb < minb) minb = absb ;
   }
printf("\nL_2 norm of RHS: ||b|| = %LG", normbee) ; 
printf("\nMin and max of RHS: min-b = %LG  max-b = %LG", minb, maxb) ; 

if (rhs||sol) normsol = sqrt(dotc(soln,soln)) ; 
// for calculating relative errors  
// p = r = DKSWP(A,b,x) - x: 
if (comp) dkswpc(x, r) ;
else dkswp(x, r) ;
for (j=1; j<=nnr; j++)
   p[j] = r[j] =  r[j] - x[j] ;
rsq = dotc(r, r) ;
printf("\nTotal setup time: %g sec.\n", (clock()-start)/cps) ;
start = clock() ;

// CG iterations: 

for (it=1; it<=maxiter; it++)
   { // q = p - DKSWP(A,0,p,omega): 
     if (comp) dkswp0c(p, q) ;
     else dkswp0(p, q) ;
     for (j=1; j<=nnr; j++) q[j]=p[j]-q[j] ;
     dot = dotc(p,q) ; 
     // avoid NAN situations long before they occur:  
     if (fabs(dot) < SMALL || fabs(rsq) < SMALL) 
        { printf("\n\nCGMN terminated at iteration %d to avoid useless \
iterations\nleading to NAN results. Accuracy cannot be improved further.\n\
(Problematic variables: dot = %LG, rsq = %LG)", it, dot, rsq) ; 
          break ; 
        } 
     alpha = rsq / dot ; 
     for (j=1; j<=nnr; j++) 
	{ x[j] += alpha*p[j] ;
	  r[j] -= alpha*q[j] ;
	}
     rsq1 = dotc(r,r) ;
     beta = rsq1 / rsq ;
     for (j=1; j<=nnr; j++)
        p[j] = r[j] + beta*p[j] ;
     rsq  = rsq1 ;

// test for printing intermediate results: 
     if (IFITER) 
	{ matvec(x, y) ; // sets y = A*x  
          resid = 0. ;
          for (j=1; j<=nnr; j++)
	     resid += (y[j]-bee[j])*(y[j]-bee[j]) ; 
          resid = sqrt(resid) ;
	  relres = resid/normbee ;
	  if (rhs || sol) // calculate relative error 
	     { relerr = 0. ;
	       for (j=1; j<=nnc; j++)
	          { if (col) 
                       err = fabs(x[j]/colnorm[j]-soln[j]) ;
                    else err = fabs(x[j]-soln[j]) ;
		    relerr += err*err ;
                  }
               relerr = sqrt(relerr)/normsol ; 
               printf("\nit = %5d  resid = %11.6LE  relres = %11.6LE  \
relerr = %11.6LE", it, resid, relres, relerr) ;
               // printf("\nrsq = %LE dot = %LE", rsq, dot) ;  

	     }
	  else 
	       printf("\nit = %5d  resid = %11.6LE  relres = %11.6LE",\
	              it, resid, relres) ; 
          if (it>=100) fflush(stdout) ; 
          if (relres <= eps) break; 
	}
   } 

// end of CG iterations  

LINE ;
timet = (clock()-start)/cps ; 

if (col) 
   for (j=1; j<=nnc; j++) 
      x[j] /= colnorm[j] ; 

matvec(x, y) ; 
maxres = -1. ;
for (j=1; j<=nnr; j++)
   { resid = fabs(y[j] - bee[j]) ;
     if (resid>maxres)
        { maxres=resid; indres=j; } 
   } 

if (rhs||sol) 
   { maxerr = -1. ;
     for (j=1; j<=nnc; j++) 
        { err = fabs(x[j]-soln[j]) ;
          if (err>maxerr)
             { maxerr=err; inderr=j; }
        }
     printf("\nmaxres = %LE at j = %d, maxerr = %LE at j = %d",\
     maxres, indres, maxerr, inderr) ;
   }
else printf("\nmaxres = %LE at j = %d", maxres, indres) ;
printf("\nmaxres/maxRHS = %LG", maxres/maxb) ; 

if (it > maxiter) it = maxiter ; 
printf("\nTime for %d CGMN iterations: %g sec.\n\n", it, timet) ;
// printf("\ncol = %d\n", col) ;  

free(p) ;
free(q) ;
free(r) ;
free(y) ;

} // end of cgmn  

